"""Utility functions."""
from datetime import datetime
from os import listdir
from pathlib import Path
from types import FrameType
from typing import TypeVar, cast
from xml.etree.ElementTree import Element
import logging
import re
import sys
import uuid
import zipfile

from dateutil.parser import isoparse
from loguru import logger

# isort: off
from .constants import (
    FEED_NAMESPACES, FEED_PROPERTIES_TAG, FEED_SUMMARY_TAG, FEED_TITLE_TAG,
    METADATA_DESCRIPTION_TAG, METADATA_DOCS_URL_TAG, METADATA_DOWNLOAD_COUNT_TAG,
    METADATA_GALLERY_DETAILS_URL_TAG, METADATA_IS_APPROVED_TAG,
    METADATA_IS_DOWNLOAD_CACHE_AVAILABLE_TAG, METADATA_LICENSE_URL_TAG,
    METADATA_PACKAGE_APPROVED_DATE_TAG, METADATA_PACKAGE_SOURCE_URL_TAG,
    METADATA_PACKAGE_TEST_RESULT_STATUS_DATE_TAG, METADATA_PACKAGE_TEST_RESULT_STATUS_TAG,
    METADATA_PROJECT_URL_TAG, METADATA_PUBLISHED_TAG, METADATA_RELEASE_NOTES_TAG, METADATA_TAGS_TAG,
    METADATA_VERSION_DOWNLOAD_COUNT_TAG, METADATA_VERSION_TAG)
# isort: on
from .typing import SearchResult, TestingStatus

__all__ = ('append_dir_to_zip_recursive', 'entry_to_search_result', 'generate_unique_id',
           'setup_logging', 'tag_text_or')

T = TypeVar('T')
utils_logger = logging.getLogger(__name__)


class InterceptHandler(logging.Handler):  # pragma: no cover
    """Intercept handler taken from Loguru's documentation."""
    def emit(self, record: logging.LogRecord) -> None:
        level: str | int
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno
        # Find caller from where originated the logged message
        frame: FrameType | None = logging.currentframe()
        depth = 2
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def setup_log_intercept_handler() -> None:  # pragma: no cover
    """Sets up Loguru to intercept records from the logging module."""
    logging.basicConfig(handlers=(InterceptHandler(),), level=0)


def setup_logging(debug: bool | None = False) -> None:
    """Shared function to enable logging."""
    if debug:  # pragma: no cover
        setup_log_intercept_handler()
        logger.enable('')
    else:
        logger.configure(handlers=(dict(
            format='<level>{message}</level>',
            level='INFO',
            sink=sys.stderr,
        ),))


def generate_unique_id() -> str:
    """Generates a unique ID for elements in ``_rels/.rels``."""
    return f'R{str(uuid.uuid4()).replace("-", "")}'.upper()


def append_dir_to_zip_recursive(root: Path, z: zipfile.ZipFile) -> None:
    """Appends a directory recursively to a zip file."""
    for item in listdir(root):
        if item.endswith('.nupkg'):
            continue
        abs_item = root / item
        if abs_item.is_dir():
            utils_logger.debug(f'Recursing into {abs_item}')
            append_dir_to_zip_recursive(abs_item, z)
        else:
            utils_logger.debug(f'Adding {abs_item}')
            z.write(abs_item)


class InvalidEntryError(ValueError):
    """Thrown when an ``<entry>`` is invalid."""


def parse_boolean_tag(tag: Element | None) -> bool:
    return ((tag.text or 'false') if tag is not None else 'false') == 'true'


def parse_iso_date_tag(tag: Element | None) -> datetime | None:
    return isoparse(tag.text) if tag is not None and tag.text else None


def parse_int_tag(tag: Element | None, default: int = 0) -> int:
    try:
        return int(tag.text) if tag is not None and tag.text else default
    except ValueError:
        return default


def tag_text_or(tag: Element | None, default: str | None = None) -> str | None:
    """Return text from a tag or the default value specified."""
    return tag.text if tag is not None and tag.text else default


def entry_to_search_result(entry: Element, ns: dict[str, str] = FEED_NAMESPACES) -> SearchResult:
    """Convert an ``<entry>`` to a ``SearchResult`` dict."""
    metadata = entry.find(FEED_PROPERTIES_TAG, ns)
    if not metadata:
        raise InvalidEntryError()
    return SearchResult(
        approval_date=parse_iso_date_tag(metadata.find(METADATA_PACKAGE_APPROVED_DATE_TAG, ns)),
        description=tag_text_or(metadata.find(METADATA_DESCRIPTION_TAG, ns)),
        documentation_uri=tag_text_or(metadata.find(METADATA_DOCS_URL_TAG, ns)),
        is_approved=parse_boolean_tag(metadata.find(METADATA_IS_APPROVED_TAG, ns)),
        is_cached=parse_boolean_tag(metadata.find(METADATA_IS_DOWNLOAD_CACHE_AVAILABLE_TAG)),
        license=tag_text_or(metadata.find(METADATA_LICENSE_URL_TAG, ns)),
        num_downloads=parse_int_tag(metadata.find(METADATA_DOWNLOAD_COUNT_TAG, ns)),
        num_version_downloads=parse_int_tag(metadata.find(METADATA_VERSION_DOWNLOAD_COUNT_TAG, ns)),
        package_src_uri=tag_text_or(metadata.find(METADATA_PACKAGE_SOURCE_URL_TAG, ns)),
        package_url=tag_text_or(metadata.find(METADATA_GALLERY_DETAILS_URL_TAG, ns)),
        publish_date=parse_iso_date_tag(metadata.find(METADATA_PUBLISHED_TAG, ns)),
        release_notes_uri=tag_text_or(metadata.find(METADATA_RELEASE_NOTES_TAG, ns)),
        site=tag_text_or(metadata.find(METADATA_PROJECT_URL_TAG, ns)),
        summary=tag_text_or(entry.find(FEED_SUMMARY_TAG, ns)),
        tags=re.split(r'\s+', cast(str, tag_text_or(metadata.find(METADATA_TAGS_TAG, ns), ''))),
        testing_status=cast(
            TestingStatus,
            tag_text_or(metadata.find(METADATA_PACKAGE_TEST_RESULT_STATUS_TAG, ns), 'Failing')),
        testing_date=parse_iso_date_tag(
            metadata.find(METADATA_PACKAGE_TEST_RESULT_STATUS_DATE_TAG, ns)),
        title=tag_text_or(entry.find(FEED_TITLE_TAG, ns)),
        version=tag_text_or(metadata.find(METADATA_VERSION_TAG, ns)))



from datetime import datetime
from typing import cast

import click

from ..client import ChocolateyClient
from ..config import read_all
from ..templates import ALL_VERSIONS_SEARCH_RESULT_TEMPLATE, SEARCH_RESULT_TEMPLATE
from ..utils import setup_logging

__all__ = ('search',)


@click.command()
@click.option('--by-id-only',
              is_flag=True,
              help='Only return packages where the id contains the search filter.')
@click.option('--by-tag-only',
              '--by-tags-only',
              is_flag=True,
              help='Only return packages where the search filter matches on the tags.')
@click.option('--id-starts-with',
              is_flag=True,
              help='Only return packages where the id starts with the search filter.')
@click.option('--idonly',
              '--id-only',
              'id_only',
              help='Only return Package IDs in the list results.',
              is_flag=True)
@click.option('--order-by-popularity', is_flag=True, help='Sort by package results by popularity.')
@click.option('--page', type=int, help='The "page" of results to return.', default=None)
@click.option('--page-size',
              type=int,
              help='Amount of packages to return in each page of results.',
              default=30)
@click.option('--pre',
              '--prerelease',
              'include_prerelease',
              help='Include prereleases.',
              is_flag=True)
@click.option('-a',
              '--all',
              '--allversions',
              '--all-versions',
              'all_versions',
              is_flag=True,
              help='Include results from all versions.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug logging.')
@click.option('-e', '--exact', is_flag=True, help='Only return packages with this exact name.')
@click.option('-p', '--password', help='Password.')
@click.option('-s', '--source', help='Source to search.', metavar='URL')
@click.option('-u', '--user', help='User name.')
@click.argument('terms')
def search(
    all_versions: bool,
    by_id_only: bool,
    by_tag_only: bool,
    debug: bool,
    exact: bool,
    id_only: bool,
    id_starts_with: bool,
    include_prerelease: bool,
    order_by_popularity: bool,
    page: int | None,
    page_size: bool,
    password: str,
    source: str,
    terms: str,
    user: str,
) -> None:
    """Search a source."""
    setup_logging(debug)
    config, api_keys = read_all()
    for result in ChocolateyClient(config, api_keys).search(
            terms,
            auth=(user, password) if user and password else None,
            all_versions=all_versions,
            by_id_only=by_id_only,
            by_tag_only=by_tag_only,
            debug=debug,
            exact=exact,
            id_starts_with=id_starts_with,
            include_prerelease=include_prerelease,
            order_by_popularity=order_by_popularity,
            page=page,
            page_size=page_size,
            source=source):
        if id_only:
            click.echo(result['title'])
        elif all_versions:
            substitutes = cast(dict[str, str | bool | None | datetime], result)
            substitutes.update(
                cached_state='Downloads cached for licensed users' if result['is_cached'] else '',
                state='[Approved]' if result['is_approved'] else '')
            click.echo(ALL_VERSIONS_SEARCH_RESULT_TEMPLATE.safe_substitute(**substitutes))
        else:
            substitutes = cast(dict[str, str | bool | None | datetime], result)
            substitutes.update(
                cached_state='Downloads cached for licensed users' if result['is_cached'] else '',
                state='[Approved]' if result['is_approved'] else '',
                tags=' '.join(result['tags']))
            substitutes['testing_status'] = f"{result['testing_status']}"
            if result['testing_date']:
                substitutes[
                    'testing_status'] += f" on {result['testing_date']}"  # type: ignore[assignment,operator]  # noqa: E501
            for key, val in substitutes.items():
                if val is None:
                    match key:
                        case 'title':
                            substitutes[key] = 'no title?'
                        case 'version':
                            substitutes[key] = 'no version?'
                        case _:
                            substitutes[key] = ''
            click.echo(SEARCH_RESULT_TEMPLATE.safe_substitute(**substitutes))



from .apikey import apikey
from .config import config
from .new import new
from .pack import pack
from .push import push
from .search import search

__all__ = ('apikey', 'config', 'new', 'pack', 'push', 'search')



"""Root command."""
import click

from choco.commands import apikey, config, new, pack, push, search


@click.group()
def main() -> None:
    """Minimal choco command."""


main.add_command(apikey, 'apikey')
main.add_command(config, 'config')
main.add_command(new, 'new')
main.add_command(pack, 'pack')
main.add_command(push, 'push')
main.add_command(search, 'search')



import click

from ..packaging import new_package


@click.command()
@click.argument('name')
def new(name: str) -> None:
    """Create a new package."""
    try:
        new_package(name)
    except (FileExistsError, ValueError) as e:
        click.secho(str(e), err=True)
        raise click.Abort() from e



"""Packaging-related functions."""
from datetime import datetime
from os import chdir
from pathlib import Path
import glob
import hashlib
import re
import zipfile

from defusedxml.ElementTree import parse as parse_xml

from .constants import (CHOCOLATEY_UNINSTALL_PS1, CONTENT_TYPES_XML, NUSPEC_FIELD_AUTHORS,
                        NUSPEC_FIELD_DESCRIPTION, NUSPEC_FIELD_ID, NUSPEC_FIELD_TAGS,
                        NUSPEC_FIELD_VERSION, VALID_NAME_RE)
from .templates import (CHOCOLATEY_INSTALL_PS1_TEMPLATE, NUSPEC_TEMPLATE, PSMDCP_XML_TEMPLATE,
                        RELS_XML_TEMPLATE)
from .utils import append_dir_to_zip_recursive, generate_unique_id, tag_text_or

__all__ = ('is_valid_package_name', 'new_package')


def is_valid_package_name(value: str) -> bool:
    """Check if a package name is valid."""
    return bool(re.match(VALID_NAME_RE, value))


def new_package(name: str) -> Path:
    """Scaffolding to create a new package."""
    if not is_valid_package_name(name):
        raise ValueError(f'Invalid package name. Name must match "{VALID_NAME_RE}".')
    if Path(name).exists():
        raise FileExistsError('Directory already exists.')
    p_name = Path(name)
    p_name.mkdir()
    with Path(p_name / f'{name}.nuspec').open('w') as f:
        f.write(NUSPEC_TEMPLATE.safe_substitute(package_id=name))
    tools = p_name / 'tools'
    tools.mkdir()
    with Path(tools / 'chocolateyInstall.ps1').open('w') as f:
        f.write(CHOCOLATEY_INSTALL_PS1_TEMPLATE.safe_substitute(package_id=name))
    with Path(tools / 'chocolateyUninstall.ps1').open('w') as f:
        f.write(CHOCOLATEY_UNINSTALL_PS1)
    return p_name


def pack(work_dir: str = '.') -> zipfile.ZipFile:
    """Pack a package directory for distribution (create a nupkg)."""
    if not (nuspecs := glob.glob('*.nuspec', root_dir=work_dir)):
        raise FileNotFoundError('No nuspec files found.')
    if len(nuspecs) > 1:
        raise ValueError(f'Only one nuspec file should be present in {work_dir}.')
    root = parse_xml(Path(work_dir) / nuspecs[0]).getroot()
    package_id = tag_text_or(root.find(NUSPEC_FIELD_ID))
    version = tag_text_or(root.find(NUSPEC_FIELD_VERSION))
    sha = hashlib.sha1()
    sha.update(f'{package_id}{version}{datetime.now()}'.encode())
    psmdcp_filename = f'{sha.hexdigest()}.psmdcp'
    with zipfile.ZipFile(f'test-{package_id}.{version}.nupkg', 'w') as z:
        chdir(work_dir)
        append_dir_to_zip_recursive(Path('.'), z)
        z.writestr('[Content_Types].xml', CONTENT_TYPES_XML)
        z.writestr(
            '_rels/.rels',
            RELS_XML_TEMPLATE.safe_substitute(nuspec_id=generate_unique_id(),
                                              package_id=package_id,
                                              psmdcp_filename=psmdcp_filename,
                                              psmdcp_id=generate_unique_id()))
        z.writestr(
            f'package/services/metadata/core-properties/{psmdcp_filename}',
            PSMDCP_XML_TEMPLATE.safe_substitute(
                creator=tag_text_or(root.find(NUSPEC_FIELD_AUTHORS)),
                description=tag_text_or(root.find(NUSPEC_FIELD_DESCRIPTION)),
                keywords=tag_text_or(root.find(NUSPEC_FIELD_TAGS)),
                package_id=package_id,
                version=version))
        return z



"""Template instances."""
import string

__all__ = ('CHOCOLATEY_INSTALL_PS1_TEMPLATE', 'NUSPEC_TEMPLATE', 'PSMDCP_XML_TEMPLATE',
           'RELS_XML_TEMPLATE', 'SEARCH_RESULT_TEMPLATE')

RELS_XML_TEMPLATE = string.Template('''<?xml version="1.0" encoding="utf-8" ?>
<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships">
  <Relationship
    Type="http://schemas.microsoft.com/packaging/2010/07/manifest"
    Target="/${package_id}.nuspec"
    Id="${nuspec_id}"
  />
    <Relationship
    Type="http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties"
    Target="/package/services/metadata/core-properties/${psmdcp_filename}"
    Id="${psmdcp_id}"
  />
</Relationships>\n''')
"""Used in construction of ``_rels/.rels`` file.

:meta hide-value:"""
last_modified_by = ('choco, Version=2.2.2.0, Culture=neutral, PublicKeyToken=79d02ea9cad655eb;'
                    'Microsoft Windows NT 10.0.22621.0;.NET Framework 4.8')
PSMDCP_XML_TEMPLATE = string.Template('''<?xml version="1.0" encoding="utf-8" ?>
<coreProperties xmlns:dc="http://purl.org/dc/elements/1.1/"
  xmlns:dcterms="http://purl.org/dc/terms/"
  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
  xmlns="http://schemas.openxmlformats.org/package/2006/metadata/core-properties">
  <dc:creator>${creator}</dc:creator>
  <dc:description>${description}</dc:description>
  <dc:identifier>${package_id}</dc:identifier>
  <version>${version}</version>
  <keywords>${keywords}</keywords>
  <lastModifiedBy>''' + last_modified_by + '''</lastModifiedBy>
</coreProperties>
\n''')  #: :meta hide-value:
"""Used in construction of ``packages/services/metadata/core-properties/{id}.psmdcp``.

:meta hide-value:"""
NUSPEC_TEMPLATE = string.Template('''<?xml version="1.0" encoding="utf-8"?>
<package xmlns="http://schemas.microsoft.com/packaging/2010/07/nuspec.xsd">
  <metadata>
    <id>${package_id}</id>
    <version>VERSION</version>
    <title>PACKAGE_NAME (Install)</title>
    <authors>AUTHORS</authors>
    <owners>NUGET_MAKER</owners>
    <requireLicenseAcceptance>false</requireLicenseAcceptance>
    <projectUrl>https://a-url</projectUrl>
    <description>DESCRIPTION</description>
    <summary>SUMMARY</summary>
    <tags>tag1 tag2</tags>
    <packageSourceUrl>https://a-url-can-be-same-as-project</packageSourceUrl>
  </metadata>
</package>\n''')
"""Used in construction of the Nuspec file.

:meta hide-value:"""
CHOCOLATEY_INSTALL_PS1_TEMPLATE = string.Template('''$$ErrorActionPreference = 'Stop'
$$packageName = '${package_id}'
$$${package_id}Version    = '1.0'
$$toolsDir   = "$$(Split-Path -parent $$MyInvocation.MyCommand.Definition)"
$$packageArgs = @{
  checksum      = ''
  checksumType  = 'sha256'
  packageName   = $$packageName
  unzipLocation = $$toolsDir
  url           = 'https://somewhere/${package_id}.zip'
}
# https://chocolatey.org/docs/helpers-install-chocolatey-zip-package
Install-ChocolateyZipPackage @packageArgs
## Unzips a file to the specified location - auto overwrites existing content
## - https://chocolatey.org/docs/helpers-get-chocolatey-unzip
#Get-ChocolateyUnzip @packageArgs\n''')
"""Chocolatey PowerShell install template.

:meta hide-value:"""

SEARCH_RESULT_TEMPLATE = string.Template('''${title} ${version} ${state} ${cached_state}
 Title: ${title} | Published: ${publish_date}
 Package approved as a trusted package on ${approval_date}.
 Package testing status: ${testing_status}.
 Number of Downloads: ${num_downloads} | Downloads for this version: ${num_version_downloads}
 Package url ${package_url}
 Chocolatey Package Source: ${package_src_uri}
 Tags:${tags}
 Software Site: ${site}
 Software License: ${license}
 Documentation: ${documentation_uri}
 Summary: ${summary}
 Description: ${description}
 Release Notes: ${release_notes_uri}
''')
"""Search result template.

:meta hide-value:"""
ALL_VERSIONS_SEARCH_RESULT_TEMPLATE = string.Template(
    '${title} ${version} ${state} ${cached_state}')
"""Search result template when ``--all-versions`` is used.

:meta hide-value:"""



import click

from ..config import read_api_keys, write_api_keys
from ..constants import PYCHOCO_API_KEYS_TOML_PATH

__all__ = ('apikey',)


@click.command('list')
@click.option('-p',
              '--path',
              default=PYCHOCO_API_KEYS_TOML_PATH,
              help='Storage file.',
              metavar='PATH',
              type=click.Path(dir_okay=False, resolve_path=True))
def list_(path: str | None) -> None:
    """Lists sources associated with API keys. Does not display the API key values."""
    try:
        keys = read_api_keys(path)
    except FileNotFoundError as e:
        raise click.Abort() from e
    for key in keys.keys():
        click.echo(key)


@click.command()
@click.option('-k', '--key', required=True, help='API key.', metavar='API_KEY')
@click.option('-s', '--source', required=True, help='Source URI.', metavar='SOURCE_URL')
@click.option('-p',
              '--path',
              default=PYCHOCO_API_KEYS_TOML_PATH,
              help='Storage file.',
              metavar='PATH',
              type=click.Path(dir_okay=False, resolve_path=True))
def add(key: str, source: str, path: str | None) -> None:
    """Add an API key for a source."""
    try:
        keys = read_api_keys(path)
    except FileNotFoundError:
        keys = {}
    keys[source] = key
    write_api_keys(keys)


@click.group()
def apikey() -> None:
    """Manage API keys."""


apikey.add_command(add, 'add')
apikey.add_command(list_, 'list')



"""Client."""
from os.path import basename
from typing import Iterator

from defusedxml.ElementTree import fromstring as parse_xml
from loguru import logger
from requests.auth import HTTPBasicAuth
import requests

from .constants import (DEFAULT_CONFIG, DEFAULT_PUSH_SOURCE, FEED_ENTRY_TAG, FEED_ID_TAG,
                        FEED_NAMESPACES, NUGET_API_KEY_HTTP_HEADER)
from .typing import Config, ConfigKey, SearchResult
from .utils import InvalidEntryError, entry_to_search_result, tag_text_or

#: Authentication details.
Auth = tuple[str, str] | HTTPBasicAuth


class ChocolateyClient:
    """
    Chocolatey client. Anything that involves interactions with the server will be in this class.
    """
    def __init__(self,
                 config: Config = DEFAULT_CONFIG,
                 api_keys: dict[str, str] | None = None) -> None:
        self.config = config
        self.api_keys = api_keys or {}
        self.session = requests.Session()

    def get_keys_available(self) -> Iterator[str]:
        """Return an iterator of the sources that have keys."""
        yield from self.api_keys.keys()

    def add_key(self, key: str, source: str) -> None:
        """Add an API key for a source."""
        self.api_keys[source] = key

    def config_set(self, name: ConfigKey, value: str) -> None:
        """Set a configuration value."""
        self.config['pychoco'][name] = value

    def get_default_push_source(self) -> str:
        """Gets the default push source URL."""
        return (self.config['pychoco'].get('defaultPushSource', DEFAULT_PUSH_SOURCE)
                or DEFAULT_PUSH_SOURCE)

    def update_api_key_header(self, source: str | None = None) -> None:
        """
        Updates the shared session API key header. If one does not exist, the existing header is
        removed if it is present.
        """
        source = source or self.get_default_push_source()
        try:
            self.session.headers.update({NUGET_API_KEY_HTTP_HEADER: self.api_keys[source]})
        except KeyError:
            try:
                del self.session.headers[NUGET_API_KEY_HTTP_HEADER]
            except KeyError:
                pass

    def push(self, package: str, source: str | None = None, auth: Auth | None = None) -> None:
        """Push a package to a source."""
        source = source or self.get_default_push_source()
        self.update_api_key_header(source)
        api_v2 = self._get_api_v2(source)
        with open(package, 'rb') as f:
            r = self.session.put(f'{source}{api_v2}/package/',
                                 auth=auth,
                                 files={basename(package): f})
            r.raise_for_status()

    def _get_api_v2(self, source: str | None = None) -> str:
        source = source or self.get_default_push_source()
        return '/api/v2' if '/api/v2' not in source else ''

    def search(self,
               terms: str,
               *,
               auth: Auth | None = None,
               all_versions: bool = False,
               by_id_only: bool = False,
               by_tag_only: bool = False,
               debug: bool = False,
               exact: bool = False,
               id_starts_with: bool = False,
               include_prerelease: bool = False,
               order_by_popularity: bool = False,
               page: int | None = None,
               page_size: bool = False,
               source: str | None = None) -> Iterator[SearchResult]:
        """Search packages. Returns a deduplicated iterator of results."""
        self.update_api_key_header(source)
        ns = FEED_NAMESPACES
        api_v2 = self._get_api_v2(source)
        source = source or self.get_default_push_source()
        searches: list[requests.Request] = []
        order_by = 'Id' if not order_by_popularity else 'DownloadCount desc,Id'
        if (not all_versions and not by_id_only and not by_tag_only and not exact
                and not id_starts_with):  # Default case
            searches.append(
                requests.Request(
                    'GET',
                    f'{source}/Packages()',
                    params={
                        '$filter':
                            f"((((Id ne null) and substringof('{terms}',tolower(Id))) or "
                            f"((Description ne null) and substringof('{terms}',"
                            'tolower(Description))))'
                            f" or ((Tags ne null) and substringof(' {terms} ',tolower(Tags)))) "
                            "and IsLatestVersion",
                        '$orderby': order_by,
                        '$skip': str(0 if page is None else page * page_size),
                        '$top': str(page_size)
                    }))
        else:
            if all_versions:
                # FIXME In both cases, pagination is not implemented. For non-exact, multiple
                # requests to FindPackagesById() are necessary with every package found. For exact,
                # $skiptoken must be used to paginate.
                searches.append(
                    requests.
                    Request('GET',
                            f'{source}{api_v2}/Search()',
                            params={
                                '$filter': 'IsLatestVersion',
                                '$orderby': order_by,
                                'includePrerelease': 'true' if include_prerelease else 'false',
                                'searchTerm': f"'{terms}'",
                                'targetFramework': "''"
                            }) if not exact else requests.
                    Request('GET', f'{source}/FindPackagesById()', params={'id': f"'{terms}'"}))
            elif exact:
                searches.append(
                    requests.Request(
                        'GET',
                        f'{source}/Packages()',
                        params={'$filter': f"(tolower(Id) eq '{terms}') and IsLatestVersion"}))
            if by_id_only:
                searches.append(
                    requests.Request(
                        'GET',
                        f'{source}/Packages()',
                        params={
                            '$filter': f"(substringof('{terms}',tolower(Id))) and IsLatestVersion",
                            '$orderby': order_by,
                            '$skip': str(0 if page is None else page * page_size),
                            '$top': str(page_size),
                            'includePrerelease': 'true' if include_prerelease else 'false',
                            'searchTerm': f"'{terms}'",
                            'targetFramework': "''"
                        }))
            if by_tag_only:
                searches.append(
                    requests.Request(
                        'GET',
                        f'{source}/Packages()',
                        params={
                            '$filter': f"(substringof('{terms}',Tags)) and IsLatestVersion",
                            '$orderby': order_by,
                            '$skip': str(0 if page is None else page * page_size),
                            '$top': str(page_size),
                            'includePrerelease': 'true' if include_prerelease else 'false',
                            'searchTerm': f"'{terms}'",
                            'targetFramework': "''"
                        }))
            if id_starts_with:
                searches.append(
                    requests.Request(
                        'GET',
                        f'{source}{api_v2}/Search()',
                        params={
                            '$filter': f"(startswith(tolower(Id),'{terms}')) and IsLatestVersion",
                            '$orderby': 'Id',
                            '$skip': str(0 if page is None else page * page_size),
                            '$top': str(page_size),
                            'includePrerelease': 'true' if include_prerelease else 'false',
                            'searchTerm': f"'{terms}'"
                        }))
        results: dict[str, SearchResult] = {}
        for i, req in enumerate(searches):
            req.auth = auth
            req.params['semVerLevel'] = '2.0.0'
            r = self.session.send(req.prepare())
            r.raise_for_status()
            if debug:  # pragma no cover
                logger.debug(r.text)
                try:
                    with open(f'result-{i:03}.xml', 'w') as f:
                        f.write(r.text)
                except IOError:
                    pass
            root = parse_xml(r.text)
            for entry in root.findall(FEED_ENTRY_TAG, ns):
                id_url = tag_text_or(entry.find(FEED_ID_TAG, ns))
                if not id_url or id_url in results:
                    continue
                try:
                    results[id_url] = entry_to_search_result(entry, ns)
                except InvalidEntryError:
                    pass
        for result in reversed(results.values()):
            yield result



import click

from ..packaging import pack as do_pack
from ..utils import setup_logging

__all__ = ('pack',)


@click.command()
@click.option('-d', '--debug', is_flag=True, help='Enable debug logging.')
@click.argument('work_dir', type=click.Path(file_okay=False, resolve_path=True), default='.')
def pack(work_dir: str = '.', debug: bool = False) -> None:
    """Create a package file for distribution."""
    setup_logging(debug)
    try:
        do_pack(work_dir)
    except (FileNotFoundError, ValueError) as e:
        click.secho(str(e), err=True)
        raise click.Abort() from e



import click

from ..config import read_config, write_config
from ..constants import DEFAULT_CONFIG, PYCHOCO_TOML_PATH
from ..typing import ConfigKey

__all__ = ('config',)


@click.command('set')
@click.option('-n',
              '--name',
              required=True,
              type=click.Choice(['defaultPushSource']),
              help='Key to set.',
              metavar='KEY')
@click.option('-v', '--value', required=True, help='Value to set.', metavar='VALUE')
@click.option('-p',
              '--path',
              default=PYCHOCO_TOML_PATH,
              help='Storage file.',
              metavar='PATH',
              type=click.Path(dir_okay=False, resolve_path=True))
def set_(name: ConfigKey, value: str, path: str | None) -> None:
    """Set a configuration value."""
    try:
        config = read_config(path)
    except FileNotFoundError:
        config = DEFAULT_CONFIG
    config['pychoco'][name] = value
    write_config(config, path)


@click.group()
def config() -> None:
    """Manage configuration."""


config.add_command(set_, 'set')



"""Persistent configuration."""
from pathlib import Path
from typing import cast

import tomlkit

from .constants import DEFAULT_CONFIG, PYCHOCO_API_KEYS_TOML_PATH, PYCHOCO_TOML_PATH
from .typing import Config

__all__ = ('read_all', 'read_api_keys', 'read_config', 'write_api_keys', 'write_config')


def read_api_keys(path: Path | str | None = None) -> dict[str, str]:
    """Read API keys stored in given path or the default path."""
    return cast(dict[str, str], tomlkit.loads(Path(path or PYCHOCO_API_KEYS_TOML_PATH).read_text()))


def write_api_keys(api_keys: dict[str, str], path: Path | str | None = None) -> None:
    """Write API keys dictionary to the given path or the default path."""
    Path(path or PYCHOCO_API_KEYS_TOML_PATH).write_text(tomlkit.dumps(api_keys))


def read_config(path: Path | str | None = None) -> Config:
    """Read configuration from the given path or the default path."""
    return cast(Config, tomlkit.loads(Path(path or PYCHOCO_TOML_PATH).read_text()))


def write_config(config: Config, path: Path | str | None = None) -> None:
    """Write configuration to the given path or the default path."""
    Path(path or PYCHOCO_TOML_PATH).write_text(tomlkit.dumps(config))


def read_all(config_path: Path | str | None = None,
             api_keys_path: Path | str | None = None) -> tuple[Config, dict[str, str] | None]:
    try:
        config = read_config(config_path)
    except FileNotFoundError:
        config = DEFAULT_CONFIG
    try:
        api_keys = read_api_keys(api_keys_path)
    except FileNotFoundError:
        api_keys = None
    return config, api_keys



from requests import HTTPError
import click

from ..client import ChocolateyClient
from ..config import read_all
from ..constants import PYCHOCO_API_KEYS_TOML_PATH, PYCHOCO_TOML_PATH


@click.command()
@click.option('-s', '--source', help='Source to upload to.', metavar='URL')
@click.option('--config-path',
              default=PYCHOCO_TOML_PATH,
              help='Configuration storage file.',
              metavar='PATH',
              type=click.Path(dir_okay=False, resolve_path=True, readable=True))
@click.option('--api-keys-path',
              default=PYCHOCO_API_KEYS_TOML_PATH,
              help='API keys storage file.',
              metavar='PATH',
              type=click.Path(dir_okay=False, resolve_path=True, readable=True))
@click.argument('package', type=click.Path(dir_okay=False, readable=True))
def push(package: str,
         source: str | None = None,
         config_path: str | None = None,
         api_keys_path: str | None = None) -> None:
    """Push a package to a source."""
    config, api_keys = read_all(config_path, api_keys_path)
    try:
        ChocolateyClient(config, api_keys).push(package, source)
    except HTTPError as e:
        click.echo(e.response)
        raise click.Abort() from e



"""Constants."""
from pathlib import Path
from typing import Final

from xdg.BaseDirectory import xdg_config_home

from .typing import Config

__all__ = ('CHOCOLATEY_UNINSTALL_PS1', 'CONTENT_TYPES_XML', 'FEED_AUTHOR_TAG', 'FEED_ENTRY_TAG',
           'FEED_ID_TAG', 'FEED_NAMESPACES', 'FEED_SUMMARY_TAG', 'FEED_TITLE_TAG',
           'FEED_UPDATED_TAG', 'METADATA_DESCRIPTION_TAG', 'METADATA_DOCS_URL_TAG',
           'METADATA_DOWNLOAD_COUNT_TAG', 'METADATA_GALLERY_DETAILS_URL_TAG',
           'METADATA_IS_APPROVED_TAG', 'METADATA_IS_DOWNLOAD_CACHE_AVAILABLE_TAG',
           'METADATA_LICENSE_URL_TAG', 'METADATA_PACKAGE_APPROVED_DATE_TAG',
           'METADATA_PACKAGE_SOURCE_URL_TAG', 'METADATA_PACKAGE_STATUS_TAG',
           'METADATA_PACKAGE_TEST_RESULT_STATUS_DATE_TAG',
           'METADATA_PACKAGE_TEST_RESULT_STATUS_TAG', 'METADATA_PROJECT_URL_TAG',
           'METADATA_PUBLISHED_TAG', 'METADATA_RELEASE_NOTES_TAG', 'METADATA_SUMMARY_TAG',
           'METADATA_TAGS_TAG', 'METADATA_VERSION_DOWNLOAD_COUNT_TAG', 'METADATA_VERSION_TAG',
           'NUSPEC_FIELD_AUTHORS', 'NUSPEC_FIELD_DESCRIPTION', 'NUSPEC_FIELD_ID',
           'NUSPEC_FIELD_TAGS', 'NUSPEC_FIELD_VERSION', 'PYCHOCO_API_KEYS_TOML_PATH',
           'PYCHOCO_TOML_PATH', 'VALID_NAME_RE')

FEED_NAMESPACES: Final[dict[str, str]] = {
    '': 'http://www.w3.org/2005/Atom',
    'd': 'http://schemas.microsoft.com/ado/2007/08/dataservices',
    'm': 'http://schemas.microsoft.com/ado/2007/08/dataservices/metadata'
}
"""Namespaces used in parsing package feeds.

:meta hide-value:"""
FEED_AUTHOR_TAG: Final[str] = 'author'
"""Feed ``author`` tag.

:meta hide-value:"""
FEED_ENTRY_TAG: Final[str] = 'entry'
"""Feed ``entry`` tag.

:meta hide-value:"""
FEED_ID_TAG: Final[str] = 'id'
"""Feed ``id`` tag.

:meta hide-value:"""
FEED_PROPERTIES_TAG: Final[str] = 'm:properties'
"""Feed ``m:properties`` tag.

:meta hide-value:"""
FEED_SUMMARY_TAG: Final[str] = 'summary'
"""Feed ``summary`` tag.

:meta hide-value:"""
FEED_TITLE_TAG: Final[str] = 'title'
"""Feed ``title`` tag.

:meta hide-value:"""
FEED_UPDATED_TAG: Final[str] = 'updated'
"""Feed ``updated`` tag.

:meta hide-value:"""
METADATA_DESCRIPTION_TAG: Final[str] = 'd:Description'
"""Feed ``d:Description`` tag.

:meta hide-value:"""
METADATA_DOCS_URL_TAG: Final[str] = 'd:DocsUrl'
"""Feed ``d:DocsUrl`` tag.

:meta hide-value:"""
METADATA_DOWNLOAD_COUNT_TAG: Final[str] = 'd:DownloadCount'
"""Feed ``d:DownloadCount`` tag.

:meta hide-value:"""
METADATA_GALLERY_DETAILS_URL_TAG: Final[str] = 'd:GalleryDetailsUrl'
"""Feed ``d:GalleryDetailsUrl`` tag.

:meta hide-value:"""
METADATA_IS_APPROVED_TAG: Final[str] = 'd:IsApproved'
"""Feed ``d:IsApproved`` tag.

:meta hide-value:"""
METADATA_IS_DOWNLOAD_CACHE_AVAILABLE_TAG: Final[str] = 'd:IsDownloadCacheAvailable'
"""Feed ``d:IsDownloadCacheAvailable`` tag.

:meta hide-value:"""
METADATA_LICENSE_URL_TAG: Final[str] = 'd:LicenseUrl'
"""Feed ``d:LicenseUrl`` tag.

:meta hide-value:"""
METADATA_PACKAGE_APPROVED_DATE_TAG: Final[str] = 'd:PackageApprovedDate'
"""Feed ``d:PackageApprovedDate`` tag.

:meta hide-value:"""
METADATA_PACKAGE_SOURCE_URL_TAG: Final[str] = 'd:PackageSourceUrl'
"""Feed ``d:PackageSourceUrl`` tag.

:meta hide-value:"""
METADATA_PACKAGE_STATUS_TAG: Final[str] = 'd:PackageStatus'
"""Feed ``d:PackageStatus`` tag.

:meta hide-value:"""
METADATA_PACKAGE_TEST_RESULT_STATUS_DATE_TAG = 'd:PackageTestResultStatusDate'
"""Feed ``d:PackageTestResultStatusDate`` tag.

:meta hide-value:"""
METADATA_PACKAGE_TEST_RESULT_STATUS_TAG: Final[str] = 'd:PackageTestResultStatus'
"""Feed ``d:PackageTestResultStatus`` tag.

:meta hide-value:"""
METADATA_PROJECT_URL_TAG: Final[str] = 'd:ProjectUrl'
"""Feed ``d:ProjectUrl`` tag.

:meta hide-value:"""
METADATA_PUBLISHED_TAG: Final[str] = 'd:Published'
"""Feed ``d:Published`` tag.

:meta hide-value:"""
METADATA_RELEASE_NOTES_TAG: Final[str] = 'd:ReleaseNotes'
"""Feed ``d:ReleaseNotes`` tag.

:meta hide-value:"""
METADATA_SUMMARY_TAG: Final[str] = 'd:Summary'
"""Feed ``d:Summary`` tag.

:meta hide-value:"""
METADATA_TAGS_TAG: Final[str] = 'd:Tags'
"""Feed ``d:Tags`` tag.

:meta hide-value:"""
METADATA_VERSION_DOWNLOAD_COUNT_TAG: Final[str] = 'd:VersionDownloadCount'
"""Feed ``d:VersionDownloadCount`` tag.

:meta hide-value:"""
METADATA_VERSION_TAG: Final[str] = 'd:Version'
"""Feed ``d:Version`` tag.

:meta hide-value:"""

NUSPEC_XSD_URI_PREFIX: Final[str] = '{http://schemas.microsoft.com/packaging/2010/07/nuspec.xsd}'
NUSPEC_FIELD_AUTHORS: Final[str] = f'{NUSPEC_XSD_URI_PREFIX}authors'
"""Nuspec authors tag.

:meta hide-value:"""
NUSPEC_FIELD_DESCRIPTION: Final[str] = f'{NUSPEC_XSD_URI_PREFIX}description'
"""Nuspec description tag.

:meta hide-value:"""
NUSPEC_FIELD_ID: Final[str] = f'{NUSPEC_XSD_URI_PREFIX}id'
"""Nuspec id tag.

:meta hide-value:"""
NUSPEC_FIELD_TAGS: Final[str] = f'{NUSPEC_XSD_URI_PREFIX}tags'
"""Nuspec tags tag.

:meta hide-value:"""
NUSPEC_FIELD_VERSION: Final[str] = f'{NUSPEC_XSD_URI_PREFIX}version'
"""Nuspec version tag.

:meta hide-value:"""
CHOCOLATEY_UNINSTALL_PS1: Final[str] = '''$$ErrorActionPreference = 'Stop'
[array]$$key = Get-UninstallRegistryKey -SoftwareName "PACKAGE_NAME"
if ($$key.Count -eq 1) {
  $$key | ForEach-Object {
    $$packageArgs = @{
      packageName = $$env:ChocolateyPackageName
      fileType    = 'exe'
      silentArgs  = '/S'
      file        = ($$_.UninstallString -split '"')[1]
    }
    Uninstall-ChocolateyPackage @packageArgs
  }
}
elseif ($$key.Count -eq 0) {
  Write-Warning "$$packageName has already been uninstalled by other means."
}
elseif ($$key.Count -gt 1) {
  Write-Warning "$$($$key.Count) matches found!"
  Write-Warning "To prevent accidental data loss, no programs will be uninstalled."
  Write-Warning "Please alert package maintainer the following keys were matched:"
  $$key | ForEach-Object { Write-Warning "- $$($$_.DisplayName)" }
}\n'''
"""Uninstall PowerShell template.

:meta hide-value:"""
CONTENT_TYPES_XML: Final[str] = '''<?xml version="1.0" encoding="utf-8"?>
<Types xmlns="http://schemas.openxmlformats.org/package/2006/content-types">
  <Default Extension="rels"
    ContentType="application/vnd.openxmlformats-package.relationships+xml" />
  <Default Extension="psmdcp"
    ContentType="application/vnd.openxmlformats-package.core-properties+xml" />
  <Default Extension="ps1" ContentType="application/octet" />
  <Default Extension="nuspec" ContentType="application/octet" />
</Types>\n'''
"""Used in creating the ``[Content_Types].xml`` file.

:meta hide-value:"""
PYCHOCO_TOML_PATH: Final[Path] = Path(xdg_config_home) / 'pychoco/config.toml'
"""Where pychoco's main configuration file is by default.

:meta hide-value:"""
PYCHOCO_API_KEYS_TOML_PATH: Final[Path] = Path(xdg_config_home) / 'pychoco/api.toml'
"""Where pychoco's API key list is stored.

:meta hide-value:"""
VALID_NAME_RE: Final[str] = r'^[0-9a-z-]+(\.(commandline|install|portable))?$'
"""Valid name for a package regular expression.

:meta hide-value:"""
NUGET_API_KEY_HTTP_HEADER: Final[str] = 'X-NuGet-ApiKey'
"""NuGet API key HTTP header.

:meta hide-value:"""
DEFAULT_PUSH_SOURCE: Final[str] = 'https://push.chocolatey.org'
"""Default push source URI.

:meta hide-value:"""
DEFAULT_CONFIG: Final[Config] = Config(pychoco={'defaultPushSource': DEFAULT_PUSH_SOURCE})
"""Default configuration.

:meta hide-value:"""



"""Typing helpers."""
from datetime import datetime
from typing import Literal, TypeVar, TypedDict

from typing_extensions import NotRequired

__all__ = ('Config', 'ConfigKey', 'SearchResult', 'TestingStatus')

#: Available keys for ``ChocolateyClient.config_set``.
ConfigKey = Literal['defaultPushSource']
T = TypeVar('T')
#: Testing state of a package.
TestingStatus = Literal['Passing', 'Failing']


class ConfigPyChocoDict(TypedDict):
    """Inner part of the configuration."""
    defaultPushSource: NotRequired[str | None]


class Config(TypedDict):
    """Configuration dictionary after de-serialisation."""
    pychoco: ConfigPyChocoDict


class SearchResult(TypedDict):
    """Represents a package search result."""
    approval_date: datetime | None
    """Approval date."""
    description: str | None
    """Description."""
    documentation_uri: str | None
    """Documentation URL."""
    is_approved: bool
    """If the package is approved."""
    is_cached: bool
    """If the package has been cached."""
    license: str | None
    """License URL."""
    num_downloads: int
    """Total number of downloads."""
    num_version_downloads: int
    """Number of downloads for this version of the package."""
    package_src_uri: str | None
    """Nuspec source."""
    package_url: str | None
    """Package download URL."""
    publish_date: datetime | None
    """Publication date."""
    release_notes_uri: str | None
    """Release notes URL."""
    site: str | None
    """Package website."""
    summary: str | None
    """Summary."""
    tags: list[str]
    """Associated tags."""
    testing_status: TestingStatus
    """Testing status."""
    testing_date: datetime | None
    """Last testing date."""
    title: str | None
    """Unique title of the package."""
    version: str | None
    """Version."""
